#include "gdamrecorderexternal.h"
#include "../gdamserverdebug.h"
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

typedef struct _GdamRecorderOutput GdamRecorderOutput;

static GtkObjectClass* parent_class = NULL;

static void gdam_recorder_external_destroy(GtkObject* object);
static int  gdam_recorder_external_read(GdamSource* source,
                                        GdamSample* sample,
					int         max_read);

/*============================================================
 *                    GdamRecorderOutput
 *
 * One sink for the recorder.
 */
struct _GdamRecorderOutput {
	gboolean		is_paused;

	guint			id;

	/* The maximum length to extend buffer to, in samples. */
	int			max_buffer;

	GdamBuffer		buffer;

	gpointer                user_data;

	GPollFD			poll_fd;

	/* If reading from a process. */
	int			pid;

	/* A glib thread (ie a number you can use with g_source_remove),
	 * but actually for private use by start/stop_output_thread. */
	guint			output_thread;

	gboolean		has_thread;

	gboolean		had_error;
};

/* Create a pure virtual GdamRecorderOutput.
 * You must fill in the three function pointers before
 * you do anything with it. */
static GdamRecorderOutput* gdam_recorder_output_new()
{
	GdamRecorderOutput* output = g_new0(GdamRecorderOutput, 1);
	gdam_buffer_construct(&output->buffer);
	return output;
}

static gboolean    gdam_recorder_output_prepare   (gpointer         src_data,
                                                   GTimeVal*        curtime,
                                                   gint*            timeout,
		                                   gpointer         user_data)
{
	GdamRecorderOutput* output = user_data;
	g_return_val_if_fail(output != NULL, FALSE);
	return FALSE;
}

static gboolean    gdam_recorder_output_check     (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
	GdamRecorderOutput* output = user_data;
	if ((output->poll_fd.revents & G_IO_ERR) == G_IO_ERR) {
		output->had_error = TRUE;
		if (GDAM_SERVER_DEBUG_POLL)
			gdam_debug("gdam_recorder_output_check: error writing");
		return TRUE;
	}
	if ((output->poll_fd.revents & G_IO_OUT) == G_IO_OUT) {
		if (GDAM_SERVER_DEBUG_POLL)
			gdam_debug("gdam_recorder_output_check: writing");
		return TRUE;
	}
	if (GDAM_SERVER_DEBUG_POLL)
		gdam_debug("gdam_recorder_output_check: not writing");
	return FALSE;
}

static gboolean    gdam_recorder_output_dispatch  (gpointer         src_data,
                                                   GTimeVal*        curtime,
		                                   gpointer         user_data)
{
	GdamRecorderOutput* output = user_data;
	if (output->buffer.size == 0 || output->had_error) {
		g_main_remove_poll(&output->poll_fd);
		output->output_thread = 0;
		output->has_thread = FALSE;
		return FALSE;
	}
	if (!gdam_buffer_write_one(&output->buffer, output->poll_fd.fd)) {
		output->had_error = TRUE;
		g_main_remove_poll(&output->poll_fd);
		output->output_thread = 0;
		output->has_thread = FALSE;
		return FALSE;
	}
	return TRUE;
}


static GSourceFuncs output_source_funcs = {
	gdam_recorder_output_prepare,
	gdam_recorder_output_check,
	gdam_recorder_output_dispatch,
	NULL
};

static void gdam_recorder_output_start(GdamRecorderOutput* output)
{
	if (output->buffer.size <= 0)
		return;
	if (output->had_error)
		return;

	output->poll_fd.events = G_IO_OUT | G_IO_ERR;
	g_assert(! output->has_thread);
	g_return_if_fail(output->output_thread == 0);
	g_main_add_poll(&output->poll_fd, G_PRIORITY_DEFAULT);
	output->output_thread = g_source_add(G_PRIORITY_DEFAULT,
					     FALSE,
	                                     &output_source_funcs,
					     NULL,
					     output,
					     NULL);
	output->has_thread = TRUE;
}
static void gdam_recorder_output_stop(GdamRecorderOutput* output)
{
	g_assert(output->has_thread);
	g_source_remove(output->output_thread);
	g_main_remove_poll(&output->poll_fd);
	output->output_thread = 0;
	output->has_thread = FALSE;
}

static void gdam_recorder_output_destroy(GdamRecorderOutput* output)
{
	if (output->has_thread) {
		gdam_recorder_output_stop(output);
		output->has_thread = FALSE;
	}
	gdam_buffer_destruct(&output->buffer);
	close(output->poll_fd.fd);
	output->poll_fd.fd = -1;
	g_free(output);
}

static GdamRecorderOutput* gdam_recorder_output_new_fd(int fd)
{
	GdamRecorderOutput* output = gdam_recorder_output_new();
	output->poll_fd.fd = fd;
	return output;
}

/*============================================================
 *                    GdamRecorderExternal
 */
static void gdam_recorder_external_init(GdamRecorderExternal* recorder_external)
{
}

static void gdam_recorder_external_class_init(GdamRecorderExternalClass* klass);
GtkType gdam_recorder_external_get_type()
{
	static GtkType recorder_external_type = 0;
	if (!recorder_external_type) {
		static const GtkTypeInfo recorder_external_info = {
			"GdamRecorderExternal",
			sizeof(GdamRecorderExternal),
			sizeof(GdamRecorderExternalClass),
			(GtkClassInitFunc) gdam_recorder_external_class_init,
			(GtkObjectInitFunc) gdam_recorder_external_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_FILTER;
		parent_class = gtk_type_class(parent);
		recorder_external_type = gtk_type_unique(parent,
		                                 &recorder_external_info);
	}
	return recorder_external_type;
}

static int  gdam_recorder_external_read(GdamSource* source,
                                        GdamSample* sample,
					int         max_read)
{
	GdamSource* filtered = GDAM_FILTER(source)->filtered;
	GdamRecorderExternal* recorder = GDAM_RECORDER_EXTERNAL(source);
	GSList* at;
	int num_read;
	if (filtered == NULL)
		return 0;
	num_read = gdam_source_read(filtered, sample, max_read);
	if (num_read <= 0) {
		source->ended = 1;
		return num_read;
	}

	if (recorder->paused)
		return num_read;

	for (at = recorder->recorders; at != NULL; at = at->next) {
		GdamRecorderOutput* output = at->data;
		int max_append;
		int discard;

		if (output->had_error)
			continue;

		max_append = MIN(output->max_buffer, num_read);
		discard = output->buffer.size / sizeof(GdamSample)
		        + max_append
		        - output->max_buffer;
		gdam_buffer_discard(&output->buffer,
		                    discard * sizeof(GdamSample));
		gdam_buffer_append(&output->buffer,
		                   sample + num_read - max_append,
				   max_append * sizeof(GdamSample));
		if (! output->has_thread) {
		 	gdam_recorder_output_start(output);
		}
	}
	return num_read;
}

static void gdam_recorder_external_destroy(GtkObject* object)
{
	GdamRecorderExternal* recorder = GDAM_RECORDER_EXTERNAL(object);
	GSList* outputs = recorder->recorders;
	recorder->recorders = NULL;
	g_slist_foreach(outputs, (GFunc) gdam_recorder_output_destroy, NULL);
	g_slist_free(outputs);
	(*parent_class->destroy)(object);
}


guint gdam_recorder_external_add_fd     (GdamRecorderExternal* recorder,
                                         int                   fd,
					 int                   buffer_size,
					 int                   pid)
{
	GdamRecorderOutput* output;
	output = gdam_recorder_output_new_fd(fd);
	if (buffer_size > 0)
		output->max_buffer = buffer_size;
	recorder->recorders = g_slist_append(recorder->recorders, output);
	output->id = ++(recorder->last_output_id);
	output->pid = pid;
	return output->id;
}

guint gdam_recorder_external_run_command(GdamRecorderExternal* recorder,
                                         const char* const*    command,
					 int                   buffer_size)
{
	int fds[2];
	int pid;
	if (pipe(fds) < 0) {
		g_warning("error creating pipe for "
		          "recorder_external_run_command");
		return 0;
	}
	if ((pid = fork()) < 0) {
		g_warning("error forking process: %s", strerror(errno));
		return 0;
	}
	if (pid == 0) {
		close(fds[1]);
		dup2(fds[0], STDIN_FILENO);
		execvp(command[0], (char**) command);
		_exit(127);
	}
	close(fds[0]);
	return gdam_recorder_external_add_fd(recorder,
	                                     fds[1],
					     buffer_size,
					     pid);
}

static gint has_id(gconstpointer output, gconstpointer id)
{
	return ((GdamRecorderOutput*)output)->id == GPOINTER_TO_UINT(id);
}

void  gdam_recorder_external_set_pause  (GdamRecorderExternal* recorder,
                                         guint                 sink_id,
					 gboolean              pause)
{
	GdamRecorderOutput* output;
	GSList* sink = g_slist_find_custom(recorder->recorders,
	                                   GUINT_TO_POINTER(sink_id),
					   has_id);
	if (sink == NULL) {
		gdam_debug("recorder external: "
		           "error pausing sink %d: doesn't exist", sink_id);
		return;
	}
	output = sink->data;
	output->is_paused = pause;
}

void  gdam_recorder_external_kill_sink  (GdamRecorderExternal* recorder,
                                         guint                 sink_id)
{
	GSList* prev = NULL;
	GSList* at;
	GdamRecorderOutput* output = NULL;
	for (at = recorder->recorders; at != NULL; at = at->next) {
		output = at->data;
		if (output->id == sink_id)
			break;
		prev = at;
	}
	if (at == NULL) {
		gdam_debug("recorder_external_kill_sink: no such sink %d",
			sink_id);
		return;
	}
	g_return_if_fail(output != NULL);
	if (prev == NULL) {
		g_assert(at == recorder->recorders);
		recorder->recorders = at->next;
	} else {
		prev->next = at->next;
	}
	g_slist_free_1(at);

	/* Now, dispense with `output';
	 * this will remove the glib polls and sources. */
	gdam_recorder_output_destroy(output);
}

#define VERIFY_NO_OUTPUT_SPECIFIED()					\
	if (fd >= 0 || command != NULL) {				\
		gdam_debug("two output flags were specified, "		\
			 "latest was %s", arg->name);			\
		continue;						\
	}

static void gdam_recorder_external_add_output(GtkObject* object,
				              int        num_in,
                                              GdamArg*   args_in,
				              int*       num_out,
				              GdamArg**  args_out)
{
	int fd = -1;
	int i;
	guint rv = 0;
	char** command = NULL;
	int buffer_size = 128 * 1024;
	GdamRecorderExternal* recorder = GDAM_RECORDER_EXTERNAL(object);
	for (i = 0; i < num_in; i++) {
		GdamArg* arg = args_in + i;
		if (strcmp(arg->name, "file") == 0
		 && arg->type == GDAM_ARG_TYPE_STRING) {
			char* fname = GDAM_ARG_STRING(arg);
			VERIFY_NO_OUTPUT_SPECIFIED();
		 	fd = open(fname, O_WRONLY|O_APPEND|O_CREAT, 0644);
			if (fd < 0) {
				goto failed;
			}
			continue;
		}
		if (strcmp(arg->name, "truncate_file") == 0
		 && arg->type == GDAM_ARG_TYPE_STRING) {
			char* fname = GDAM_ARG_STRING(arg);
			VERIFY_NO_OUTPUT_SPECIFIED();
		 	fd = open(fname, O_WRONLY|O_TRUNC|O_CREAT, 0644);
			if (fd < 0) {
				goto failed;
			}
			continue;
		}
		if (strcmp(arg->name, "connect") == 0
		 && arg->type == GDAM_ARG_TYPE_STRING) {
			/* nc host port */
			char* spec = GDAM_ARG_STRING(arg);
			char* colon;
			VERIFY_NO_OUTPUT_SPECIFIED();
		 	command = g_new(char*, 4);
			colon = strchr(spec, ':');
			command[0] = g_strdup("nc");
			if (colon == NULL) {
				command[1] = g_strdup(spec);
				command[2] = NULL;
			} else {
				command[1] = g_new(char, colon-spec+1);
				command[2] = g_strdup(colon + 1);
				command[3] = NULL;
				memcpy(command[1], spec, colon - spec);
				command[1][colon - spec] = '\0';
			}
			continue;
		}
		if (strcmp(arg->name, "buffer_size") == 0
		 && arg->type == GDAM_ARG_TYPE_INT) {
		 	buffer_size = GDAM_ARG_INT(arg);
			continue;
		}
	}
	if (command != NULL) {
		rv = gdam_recorder_external_run_command(recorder,
		                 (const char* const*)   command,
							buffer_size);
	} else if (fd >= 0) {
		rv = gdam_recorder_external_add_fd(recorder,
		                                   fd,
						   buffer_size,
						   0);
	}
	if (rv == 0)
		goto failed;
	
	*num_out = 1;
	*args_out = g_new(GdamArg, 1);
	(*args_out)->name = g_strdup("output_id");
	(*args_out)->type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT(*args_out) = rv;
	return;

failed:
	*args_out = NULL;
	*num_out = 0;
}
static void gdam_recorder_external_remove_output(GtkObject* object,
				                 int        num_in,
                                                 GdamArg*   args_in,
				                 int*       num_out,
				                 GdamArg**  args_out)
{
	int i;
	guint output_id = 0;
	GdamRecorderExternal* recorder = GDAM_RECORDER_EXTERNAL(object);
	for (i = 0; i < num_in; i++) {
		GdamArg* arg = args_in + i;
		if (strcmp(arg->name, "output_id") == 0
		 && arg->type == GDAM_ARG_TYPE_INT) {
			output_id = GDAM_ARG_INT(arg);
			continue;
		}
		gdam_debug("gdam_recorder_external_remove_output: "
		           "argument type %s unknown", arg->name);
	}
	if (output_id == 0) {
		*args_out = NULL;
		*num_out = 0;
		return;
	}
	gdam_recorder_external_kill_sink(recorder, output_id);
	*num_out = 1;
	*args_out = g_new(GdamArg, 1);
	(*args_out)->name = g_strdup("success");
	(*args_out)->type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT(*args_out) = 1;
	return;
}


static void gdam_recorder_external_class_init(GdamRecorderExternalClass* klass)
{
	GdamSourceClass* source_class = GDAM_SOURCE_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	source_class->read = gdam_recorder_external_read;
	object_class->destroy = gdam_recorder_external_destroy;
	gdam_arg_function_register(object_class,
			           "add_output",
			           gdam_recorder_external_add_output);
	gdam_arg_function_register(object_class,
			           "remove_output",
			           gdam_recorder_external_remove_output);
}

